searchData={"items":[{"type":"module","title":"FunWithFlags","doc":"FunWithFlags, the Elixir feature flag library.\n\nThis module provides the public interface to the library and its API is\nmade of three simple methods to enable, disable and query feature flags.\n\nIn their simplest form, flags can be toggled on and off globally.\n\nMore advanced rules or \"gates\" are available, and they can be set and queried\nfor any term that implements these protocols:\n\n* The `FunWithFlags.Actor` protocol can be\nimplemented for types and structs that should have specific rules. For\nexample, in web applications it's common to use a `%User{}` struct or\nequivalent as an actor, or perhaps the current country of the request.\n\n* The `FunWithFlags.Group` protocol can be\nimplemented for types and structs that should belong to groups for which\none wants to enable and disable some flags. For example, one could implement\nthe protocol for a `%User{}` struct to identify administrators.\n\n\nSee the [Usage](/fun_with_flags/readme.html#usage) notes for a more detailed\nexplanation.","ref":"FunWithFlags.html"},{"type":"function","title":"FunWithFlags.all_flag_names/0","doc":"Returns a list of all flag names currently configured, as atoms.\n\nThis can be useful for debugging or for display purposes,\nbut it's not meant to be used at runtime. Undefined flags,\nfor example, will be considered disabled.","ref":"FunWithFlags.html#all_flag_names/0"},{"type":"function","title":"FunWithFlags.all_flags/0","doc":"Returns a list of all the flags currently configured, as data structures.\n\nThis function is provided for debugging and to build more complex\nfunctionality (e.g. it's used in the web GUI), but it is not meant to be\nused at runtime to check if a flag is enabled.\n\nTo query the value of a flag, please use the `enabled?2` function instead.","ref":"FunWithFlags.html#all_flags/0"},{"type":"function","title":"FunWithFlags.clear/2","doc":"Clears the data of a feature flag.\n\nClears the data for an entire feature flag or for a specific\nActor or Group gate. Clearing a boolean gate is not supported\nbecause a missing boolean gate is equivalent to a disabled boolean\ngate.\n\nSometimes enabling or disabling a gate is not what you want, and you\nneed to remove that gate's rules instead. For example, if you don't need\nanymore to explicitly enable or disable a flag for an actor, and the\ndefault state should be used instead, you'll want to clear the gate.\n\nIt's also possible to clear the entire flag, by not passing any option.","ref":"FunWithFlags.html#clear/2"},{"type":"function","title":"Options - FunWithFlags.clear/2","doc":"* `for_actor: an_actor` - used to clear the flag for a specific term only.\nThe value can be any term that implements the `Actor` protocol.\n* `for_group: a_group_name` - used to clear the flag for a specific group only.\n The value should be a binary or an atom (It's internally converted\nto a binary and it's stored and retrieved as a binary. Atoms are\nsupported for retro-compatibility with versions <= 0.9)\n* `boolean: true` - used to clear the boolean gate.\n* `for_percentage: true` - used to clear any percentage gate.","ref":"FunWithFlags.html#clear/2-options"},{"type":"function","title":"Examples - FunWithFlags.clear/2","doc":"iex> alias FunWithFlags.TestUser, as: User\n    iex> harry = %User{id: 1, name: \"Harry Potter\", groups: [\"wizards\", \"gryffindor\"]}\n    iex> hagrid = %User{id: 2, name: \"Rubeus Hagrid\", groups: [\"wizards\", \"gamekeeper\"]}\n    iex> dudley = %User{id: 3, name: \"Dudley Dursley\", groups: [\"muggles\"]}\n    iex> FunWithFlags.disable(:wands)\n    iex> FunWithFlags.enable(:wands, for_group: \"wizards\")\n    iex> FunWithFlags.disable(:wands, for_actor: hagrid)\n    iex>\n    iex> FunWithFlags.enabled?(:wands)\n    false\n    iex> FunWithFlags.enabled?(:wands, for: harry)\n    true\n    iex> FunWithFlags.enabled?(:wands, for: hagrid)\n    false\n    iex> FunWithFlags.enabled?(:wands, for: dudley)\n    false\n    iex>\n    iex> FunWithFlags.clear(:wands, for_actor: hagrid)\n    :ok\n    iex> FunWithFlags.enabled?(:wands, for: hagrid)\n    true\n    iex>\n    iex> FunWithFlags.clear(:wands)\n    :ok\n    iex> FunWithFlags.enabled?(:wands)\n    false\n    iex> FunWithFlags.enabled?(:wands, for: harry)\n    false\n    iex> FunWithFlags.enabled?(:wands, for: hagrid)\n    false\n    iex> FunWithFlags.enabled?(:wands, for: dudley)\n    false","ref":"FunWithFlags.html#clear/2-examples"},{"type":"function","title":"FunWithFlags.disable/2","doc":"Disables a feature flag.","ref":"FunWithFlags.html#disable/2"},{"type":"function","title":"Options - FunWithFlags.disable/2","doc":"* `:for_actor` - used to disable the flag for a specific term only.\nThe value can be any term that implements the `Actor` protocol.\n* `:for_group` - used to disable the flag for a specific group only.\n The value should be a binary or an atom (It's internally converted\nto a binary and it's stored and retrieved as a binary. Atoms are\nsupported for retro-compatibility with versions <= 0.9)\n* `:for_percentage_of` - used to disable the flag for a percentage\nof time or actors, expressed as `{:time, float}` or `{:actors, float}`,\nwhere float is in the range `0.0 < x < 1.0`.","ref":"FunWithFlags.html#disable/2-options"},{"type":"function","title":"Examples - FunWithFlags.disable/2","doc":"","ref":"FunWithFlags.html#disable/2-examples"},{"type":"function","title":"Disable globally - FunWithFlags.disable/2","doc":"iex> FunWithFlags.enable(:random_koala_gifs)\n    iex> FunWithFlags.enabled?(:random_koala_gifs)\n    true\n    iex> FunWithFlags.disable(:random_koala_gifs)\n    {:ok, false}\n    iex> FunWithFlags.enabled?(:random_koala_gifs)\n    false","ref":"FunWithFlags.html#disable/2-disable-globally"},{"type":"function","title":"Disable for an actor - FunWithFlags.disable/2","doc":"iex> FunWithFlags.enable(:spider_sense)\n    {:ok, true}\n    iex> villain = %{name: \"Venom\"}\n    iex> FunWithFlags.disable(:spider_sense, for_actor: villain)\n    {:ok, false}\n    iex> FunWithFlags.enabled?(:spider_sense)\n    true\n    iex> FunWithFlags.enabled?(:spider_sense, for: villain)\n    false","ref":"FunWithFlags.html#disable/2-disable-for-an-actor"},{"type":"function","title":"Disable for a group - FunWithFlags.disable/2","doc":"This example relies on the [reference implementation](https://github.com/tompave/fun_with_flags/blob/master/test/support/test_user.ex)\nused in the tests.\n\n    iex> alias FunWithFlags.TestUser, as: User\n    iex> harry = %User{name: \"Harry Potter\", groups: [\"wizards\", \"gryffindor\"]}\n    iex> dudley = %User{name: \"Dudley Dursley\", groups: [\"muggles\"]}\n    iex> FunWithFlags.enable(:hogwarts)\n    {:ok, true}\n    iex> FunWithFlags.disable(:hogwarts, for_group: \"muggles\")\n    {:ok, false}\n    iex> FunWithFlags.enabled?(:hogwarts)\n    true\n    iex> FunWithFlags.enabled?(:hogwarts, for: harry)\n    true\n    iex> FunWithFlags.enabled?(:hogwarts, for: dudley)\n    false","ref":"FunWithFlags.html#disable/2-disable-for-a-group"},{"type":"function","title":"Disable for a percentage of the time - FunWithFlags.disable/2","doc":"iex> FunWithFlags.clear(:random_glitch)\n    :ok\n    iex> FunWithFlags.disable(:random_glitch, for_percentage_of: {:time, 0.999999999})\n    {:ok, false}\n    iex> FunWithFlags.enabled?(:random_glitch)\n    false\n    iex> FunWithFlags.disable(:random_glitch, for_percentage_of: {:time, 0.000000001})\n    {:ok, false}\n    iex> FunWithFlags.enabled?(:random_glitch)\n    true","ref":"FunWithFlags.html#disable/2-disable-for-a-percentage-of-the-time"},{"type":"function","title":"Disable for a percentage of the actors - FunWithFlags.disable/2","doc":"iex> FunWithFlags.disable(:new_ui, for_percentage_of: {:actors, 0.3})\n    {:ok, false}","ref":"FunWithFlags.html#disable/2-disable-for-a-percentage-of-the-actors"},{"type":"function","title":"FunWithFlags.enable/2","doc":"Enables a feature flag.","ref":"FunWithFlags.html#enable/2"},{"type":"function","title":"Options - FunWithFlags.enable/2","doc":"* `:for_actor` - used to enable the flag for a specific term only.\nThe value can be any term that implements the `Actor` protocol.\n* `:for_group` - used to enable the flag for a specific group only.\nThe value should be a binary or an atom (It's internally converted\nto a binary and it's stored and retrieved as a binary. Atoms are\nsupported for retro-compatibility with versions <= 0.9)\n* `:for_percentage_of` - used to enable the flag for a percentage\nof time or actors, expressed as `{:time, float}` or `{:actors, float}`,\nwhere float is in the range `0.0 < x < 1.0`.","ref":"FunWithFlags.html#enable/2-options"},{"type":"function","title":"Examples - FunWithFlags.enable/2","doc":"","ref":"FunWithFlags.html#enable/2-examples"},{"type":"function","title":"Enable globally - FunWithFlags.enable/2","doc":"iex> FunWithFlags.enabled?(:super_shrink_ray)\n    false\n    iex> FunWithFlags.enable(:super_shrink_ray)\n    {:ok, true}\n    iex> FunWithFlags.enabled?(:super_shrink_ray)\n    true","ref":"FunWithFlags.html#enable/2-enable-globally"},{"type":"function","title":"Enable for an actor - FunWithFlags.enable/2","doc":"iex> FunWithFlags.disable(:warp_drive)\n    {:ok, false}\n    iex> FunWithFlags.enable(:warp_drive, for_actor: \"Scotty\")\n    {:ok, true}\n    iex> FunWithFlags.enabled?(:warp_drive)\n    false\n    iex> FunWithFlags.enabled?(:warp_drive, for: \"Scotty\")\n    true","ref":"FunWithFlags.html#enable/2-enable-for-an-actor"},{"type":"function","title":"Enable for a group - FunWithFlags.enable/2","doc":"This example relies on the [reference implementation](https://github.com/tompave/fun_with_flags/blob/master/test/support/test_user.ex)\nused in the tests.\n\n    iex> alias FunWithFlags.TestUser, as: User\n    iex> marty = %User{name: \"Marty McFly\", groups: [\"students\", \"time_travelers\"]}\n    iex> doc = %User{name: \"Emmet Brown\", groups: [\"scientists\", \"time_travelers\"]}\n    iex> buford = %User{name: \"Buford Tannen\", groups: [\"gunmen\", \"bandits\"]}\n    iex> FunWithFlags.enable(:delorean, for_group: \"time_travelers\")\n    {:ok, true}\n    iex> FunWithFlags.enabled?(:delorean)\n    false\n    iex> FunWithFlags.enabled?(:delorean, for: buford)\n    false\n    iex> FunWithFlags.enabled?(:delorean, for: marty)\n    true\n    iex> FunWithFlags.enabled?(:delorean, for: doc)\n    true","ref":"FunWithFlags.html#enable/2-enable-for-a-group"},{"type":"function","title":"Enable for a percentage of the time - FunWithFlags.enable/2","doc":"iex> FunWithFlags.disable(:random_glitch)\n    iex> FunWithFlags.enable(:random_glitch, for_percentage_of: {:time, 0.999999999})\n    iex> FunWithFlags.enabled?(:random_glitch)\n    true\n    iex> FunWithFlags.enable(:random_glitch, for_percentage_of: {:time, 0.000000001})\n    iex> FunWithFlags.enabled?(:random_glitch)\n    false","ref":"FunWithFlags.html#enable/2-enable-for-a-percentage-of-the-time"},{"type":"function","title":"Enable for a percentage of the actors - FunWithFlags.enable/2","doc":"This example is based on the fact that the actor score for the actor-flag pair\n`marty + :new_ui` is lower than 50%, and for the `buford + :new_ui` is higher.\n\n    iex> FunWithFlags.disable(:new_ui)\n    iex> FunWithFlags.enable(:new_ui, for_percentage_of: {:actors, 0.5})\n    iex> FunWithFlags.enabled?(:new_ui)\n    false\n    iex> alias FunWithFlags.TestUser, as: User\n    iex> marty = %User{id: 42, name: \"Marty McFly\"}\n    iex> buford = %User{id: 2, name: \"Buford Tannen\"}\n    iex> FunWithFlags.enabled?(:new_ui, for: marty)\n    true\n    iex> FunWithFlags.enabled?(:new_ui, for: buford)\n    false","ref":"FunWithFlags.html#enable/2-enable-for-a-percentage-of-the-actors"},{"type":"function","title":"FunWithFlags.enabled?/2","doc":"Checks if a flag is enabled.\n\nIt can be invoked with just the flag name, as an atom,\nto check the general status of a flag (i.e. the boolean gate).","ref":"FunWithFlags.html#enabled?/2"},{"type":"function","title":"Options - FunWithFlags.enabled?/2","doc":"* `:for` - used to provide a term for which the flag could\nhave a specific value. The passed term should implement the\n`Actor` or `Group` protocol, or both.","ref":"FunWithFlags.html#enabled?/2-options"},{"type":"function","title":"Examples - FunWithFlags.enabled?/2","doc":"This example relies on the [reference implementation](https://github.com/tompave/fun_with_flags/blob/master/test/support/test_user.ex)\nused in the tests.\n\n    iex> alias FunWithFlags.TestUser, as: User\n    iex> harry = %User{id: 1, name: \"Harry Potter\", groups: [\"wizards\", \"gryffindor\"]}\n    iex> FunWithFlags.disable(:elder_wand)\n    iex> FunWithFlags.enable(:elder_wand, for_actor: harry)\n    iex> FunWithFlags.enabled?(:elder_wand)\n    false\n    iex> FunWithFlags.enabled?(:elder_wand, for: harry)\n    true\n    iex> voldemort = %User{id: 7, name: \"Tom Riddle\", groups: [\"wizards\", \"slytherin\"]}\n    iex> FunWithFlags.enabled?(:elder_wand, for: voldemort)\n    false\n    iex> filch = %User{id: 88, name: \"Argus Filch\", groups: [\"staff\"]}\n    iex> FunWithFlags.enable(:magic_wands, for_group: \"wizards\")\n    iex> FunWithFlags.enabled?(:magic_wands, for: harry)\n    true\n    iex> FunWithFlags.enabled?(:magic_wands, for: voldemort)\n    true\n    iex> FunWithFlags.enabled?(:magic_wands, for: filch)\n    false","ref":"FunWithFlags.html#enabled?/2-examples"},{"type":"function","title":"FunWithFlags.get_flag/1","doc":"Returns a `FunWithFlags.Flag` struct for the given name, or `nil` if\nno flag is found.\n\nUseful for debugging.","ref":"FunWithFlags.html#get_flag/1"},{"type":"type","title":"FunWithFlags.options/0","doc":"","ref":"FunWithFlags.html#t:options/0"},{"type":"protocol","title":"FunWithFlags.Actor","doc":"Implement this protocol to provide actors.\n\n\nActor gates allows you to enable or disable a flag for one or more entities.\nFor example, in web applications it's common to use a `%User{}` struct or\nequivalent as an actor, or perhaps the data used to represent the current\ncountry for an HTTP request.\nThis can be useful to showcase a work-in-progress feature to someone, to\ngradually rollout a functionality by country, or to dynamically disable some\nfeatures in some contexts (e.g. a deploy introduces a critical error that\nonly happens in one specific country).\n\nActor gates take precedence over the others, both when they're enabled and\nwhen they're disabled. They can be considered as toggle overrides.\n\n\nIn order to be used as an actor, an entity must implement\nthe `FunWithFlags.Actor` protocol. This can be implemented for custom structs\nor literally any other type.","ref":"FunWithFlags.Actor.html"},{"type":"protocol","title":"Examples - FunWithFlags.Actor","doc":"This protocol is typically implemented for some application structure.\n\n    defmodule MyApp.User do\n      defstruct [:id, :name]\n    end\n\n    defimpl FunWithFlags.Actor, for: MyApp.User do\n      def id(%{id: id}) do\n        \"user:#{id}\"\n      end\n    end\n\n    bruce = %User{id: 1, name: \"Bruce\"}\n    alfred = %User{id: 2, name: \"Alfred\"}\n\n    FunWithFlags.Actor.id(bruce)\n    \"user:1\"\n    FunWithFlags.Actor.id(alfred)\n    \"user:2\"\n\n    FunWithFlags.enable(:batmobile, for_actor: bruce)\n\n\nbut it can also be implemented for the builtin types:\n\n\n    defimpl FunWithFlags.Actor, for: Map do\n      def id(%{actor_id: actor_id}) do\n        \"map:#{actor_id}\"\n      end\n\n      def id(map) do\n        map\n        |> inspect()\n        |> (&:crypto.hash(:md5, &1)).()\n        |> Base.encode16\n        |> (&\"map:#{&1}\").()\n      end\n    end\n\n\n    defimpl FunWithFlags.Actor, for: BitString do\n      def id(str) do\n        \"string:#{str}\"\n      end\n    end\n\n    FunWithFlags.Actor.id(%{actor_id: \"bar\"})\n    \"map:bar\"\n    FunWithFlags.Actor.id(%{foo: \"bar\"})\n    \"map:E0BB5BA6873E3AC34B0B6928190C1F2B\"\n    FunWithFlags.Actor.id(\"foobar\")\n    \"string:foobar\"\n\n\n    FunWithFlags.disable(:foobar, for_actor: %{actor_id: \"just a map\"})\n    FunWithFlags.enable(:foobar, for_actor: \"just a string\")\n\n\nActor identifiers must be globally unique binaries. Since supporting multiple\nkinds of actors is a common requirement, all the examples use the common\ntechnique of namespacing the IDs:\n\n\n    defimpl FunWithFlags.Actor, for: MyApp.User do\n      def id(user) do\n        \"user:#{user.id}\"\n      end\n    end\n\n    defimpl FunWithFlags.Actor, for: MyApp.Country do\n      def id(country) do\n        \"country:#{country.iso3166}\"\n      end\n    end","ref":"FunWithFlags.Actor.html#module-examples"},{"type":"function","title":"FunWithFlags.Actor.id/1","doc":"Should return a globally unique binary.","ref":"FunWithFlags.Actor.html#id/1"},{"type":"function","title":"Example - FunWithFlags.Actor.id/1","doc":"iex> FunWithFlags.Actor.id(%FunWithFlags.TestUser{id: 313})\n    \"user:313\"","ref":"FunWithFlags.Actor.html#id/1-example"},{"type":"type","title":"FunWithFlags.Actor.t/0","doc":"All the types that implement this protocol.","ref":"FunWithFlags.Actor.html#t:t/0"},{"type":"module","title":"FunWithFlags.Flag","doc":"Represents a feature flag.\n\nThis module is not meant to be used directly.","ref":"FunWithFlags.Flag.html"},{"type":"type","title":"FunWithFlags.Flag.t/0","doc":"","ref":"FunWithFlags.Flag.html#t:t/0"},{"type":"module","title":"FunWithFlags.Gate","doc":"Represents a feature flag gate, that is one of several conditions\nattached to a feature flag.\n\nThis module is not meant to be used directly.","ref":"FunWithFlags.Gate.html"},{"type":"type","title":"FunWithFlags.Gate.t/0","doc":"","ref":"FunWithFlags.Gate.html#t:t/0"},{"type":"protocol","title":"FunWithFlags.Group","doc":"Implement this protocol to provide groups.\n\nGroup gates are similar to actor gates, but they apply to a category of entities rather than specific ones. They can be toggled on or off for the _name of the group_ (as an atom) instead of a specific term.\n\nGroup gates take precedence over boolean gates but are overridden by actor gates.\n\nThe semantics to determine which entities belong to which groups are application specific.\nEntities could have an explicit list of groups they belong to, or the groups could be abstract and inferred from some other attribute. For example, an `:employee` group could comprise all `%User{}` structs with an email address matching the company domain, or an `:admin` group could be made of all users with `%User{admin: true}`.\n\nIn order to be affected by a group gate, an entity should implement the `FunWithFlags.Group` protocol. The protocol automatically falls back to a default `Any` implementation, which states that any entity belongs to no group at all. This makes it possible to safely use \"normal\" actors when querying group gates, and to implement the protocol only for structs and types for which it matters.\n\nThe protocol can be implemented for custom structs or literally any other type.\n\n\n    defmodule MyApp.User do\n      defstruct [:email, admin: false, groups: []]\n    end\n\n    defimpl FunWithFlags.Group, for: MyApp.User do\n      def in?(%{email: email}, :employee),  do: Regex.match?(~r/@mycompany.com$/, email)\n      def in?(%{admin: is_admin}, :admin),  do: !!is_admin\n      def in?(%{groups: list}, group_name), do: group_name in list\n    end\n\n    elisabeth = %User{email: \"elisabeth@mycompany.com\", admin: true, groups: [:engineering, :product]}\n    FunWithFlags.Group.in?(elisabeth, :employee)\n    true\n    FunWithFlags.Group.in?(elisabeth, :admin)\n    true\n    FunWithFlags.Group.in?(elisabeth, :engineering)\n    true\n    FunWithFlags.Group.in?(elisabeth, :marketing)\n    false\n\n    defimpl FunWithFlags.Group, for: Map do\n      def in?(%{group: group_name}, group_name), do: true\n      def in?(_, _), do: false\n    end\n\n    FunWithFlags.Group.in?(%{group: :dumb_tests}, :dumb_tests)\n    true\n\nWith the protocol implemented, actors can be used with the library functions:\n\n\n    FunWithFlags.disable(:database_access)\n    FunWithFlags.enable(:database_access, for_group: :engineering)","ref":"FunWithFlags.Group.html"},{"type":"function","title":"FunWithFlags.Group.in?/2","doc":"Should return a boolean.\n\nThe default implementation will always return `false` for\nany argument.","ref":"FunWithFlags.Group.html#in?/2"},{"type":"function","title":"Example - FunWithFlags.Group.in?/2","doc":"iex> user = %{name: \"bolo\", group: \"staff\"}\n    iex> FunWithFlags.Group.in?(data, \"staff\")\n    true\n    iex> FunWithFlags.Group.in?(data, \"superusers\")\n    false","ref":"FunWithFlags.Group.html#in?/2-example"},{"type":"type","title":"FunWithFlags.Group.t/0","doc":"All the types that implement this protocol.","ref":"FunWithFlags.Group.html#t:t/0"},{"type":"module","title":"FunWithFlags.Store.Cache","doc":"The in-memory cache for the feature flag, backed by an ETS table.\n\nThis module is not meant to be used directly, but some of its functions can be\nuseful to debug flag state.","ref":"FunWithFlags.Store.Cache.html"},{"type":"function","title":"FunWithFlags.Store.Cache.dump/0","doc":"Returns the contents of the cache ETS table, for inspection.","ref":"FunWithFlags.Store.Cache.html#dump/0"},{"type":"function","title":"FunWithFlags.Store.Cache.flush/0","doc":"Clears the cache. It will be rebuilt gradually as the public interface of the\npackage is queried.","ref":"FunWithFlags.Store.Cache.html#flush/0"},{"type":"type","title":"FunWithFlags.Store.Cache.cached_at/0","doc":"","ref":"FunWithFlags.Store.Cache.html#t:cached_at/0"},{"type":"type","title":"FunWithFlags.Store.Cache.ttl/0","doc":"","ref":"FunWithFlags.Store.Cache.html#t:ttl/0"},{"type":"behaviour","title":"FunWithFlags.Store.Persistent","doc":"A behaviour module for implementing persistence adapters.\n\nThe package ships with persistence adapters for Redis and Ecto, but you\ncan provide your own adapters by adopting this behaviour.","ref":"FunWithFlags.Store.Persistent.html"},{"type":"callback","title":"FunWithFlags.Store.Persistent.all_flag_names/0","doc":"Retrieves all the names of the persisted flags.","ref":"FunWithFlags.Store.Persistent.html#c:all_flag_names/0"},{"type":"callback","title":"FunWithFlags.Store.Persistent.all_flags/0","doc":"Retrieves all the persisted flags.","ref":"FunWithFlags.Store.Persistent.html#c:all_flags/0"},{"type":"callback","title":"FunWithFlags.Store.Persistent.delete/1","doc":"Deletes an entire flag, identified by name.","ref":"FunWithFlags.Store.Persistent.html#c:delete/1"},{"type":"callback","title":"FunWithFlags.Store.Persistent.delete/2","doc":"Deletes a gate from a flag, identified by name.","ref":"FunWithFlags.Store.Persistent.html#c:delete/2"},{"type":"callback","title":"FunWithFlags.Store.Persistent.get/1","doc":"Retrieves a flag by name.","ref":"FunWithFlags.Store.Persistent.html#c:get/1"},{"type":"callback","title":"FunWithFlags.Store.Persistent.put/2","doc":"Persists a gate for a flag, identified by name.","ref":"FunWithFlags.Store.Persistent.html#c:put/2"},{"type":"callback","title":"FunWithFlags.Store.Persistent.worker_spec/0","doc":"A persistent adapter should return either\n[a child specification](https://hexdocs.pm/elixir/Supervisor.html#module-child-specification)\nif it needs any process to be started and supervised, or `nil` if it does not.\n\nFor example, the builtin Redis persistence adapter implements this function by delegating to\n`Redix.child_spec/1` because it needs the Redix processes to work. On the other hand, the\nbuiltin Ecto adapter implements this function by returning `nil`, because the Ecto repo is\nprovided to this package by the host application, and it's assumed that the Ecto process tree\nis started and supervised somewhere else.\n\nThis custom `worker_spec/0` function is used instead of the typical `child_spec/1` function\nbecause this function can return `nil` if the adapter doesn't need to be supervised, whereas\n`child_spec/1` _must_ return a valid child spec map.","ref":"FunWithFlags.Store.Persistent.html#c:worker_spec/0"},{"type":"module","title":"FunWithFlags.Supervisor","doc":"A [Module-based supervisor](https://hexdocs.pm/elixir/Supervisor.html#module-module-based-supervisors).\n\nIt implements [`Supervisor.child_spec/1`](https://hexdocs.pm/elixir/Supervisor.html#module-child_spec-1) to describe the supervision tree for the\n`:fun_with_flags` application.\n\nThis module is used internally by the package when the `:fun_with_flags` OTP\napplication starts its own supervision tree, which is the default behavior.\nIf that is disabled, the user's host application should use this module to start\nthe supervision tree directly.\n\nThe main purpose of this API is allow the user's host application to control\nwhen FunWithFlag's supervision tree is started. This is helpful when the\npackage is configured to depend on some of the host application's modules, e.g.\nthe `Phoenix.PubSub` process ([as documented](readme.html#pubsub-adapters)).\n\nMore detailed instructions on how to configure this in an application are\navailable in the [readme](readme.html#application-start-behaviour).","ref":"FunWithFlags.Supervisor.html"},{"type":"function","title":"FunWithFlags.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"FunWithFlags.Supervisor.html#child_spec/1"},{"type":"function","title":"FunWithFlags.Supervisor.start_link/1","doc":"How to start this supervisor and its tree.\n\nThis function is referenced by the `child_spec/1` definition for this supervisor module.","ref":"FunWithFlags.Supervisor.html#start_link/1"},{"type":"exception","title":"FunWithFlags.Gate.InvalidGroupNameError","doc":"","ref":"FunWithFlags.Gate.InvalidGroupNameError.html"},{"type":"exception","title":"FunWithFlags.Gate.InvalidTargetError","doc":"","ref":"FunWithFlags.Gate.InvalidTargetError.html"},{"type":"extras","title":"FunWithFlags","doc":"# FunWithFlags\n\n[![Mix Tests](https://github.com/tompave/fun_with_flags/actions/workflows/test.yml/badge.svg?branch=master)](https://github.com/tompave/fun_with_flags/actions/workflows/test.yml?query=branch%3Amaster)\n[![Code Quality](https://github.com/tompave/fun_with_flags/actions/workflows/quality.yml/badge.svg?branch=master)](https://github.com/tompave/fun_with_flags/actions/workflows/quality.yml?query=branch%3Amaster)  \n[![Hex.pm](https://img.shields.io/hexpm/v/fun_with_flags.svg)](https://hex.pm/packages/fun_with_flags)\n[![hexdocs.pm](https://img.shields.io/badge/docs-1.12.0-brightgreen.svg)](https://hexdocs.pm/fun_with_flags/1.12.0/FunWithFlags.html)\n[![Hex.pm Downloads](https://img.shields.io/hexpm/dt/fun_with_flags)](https://hex.pm/packages/fun_with_flags)\n[![License](https://img.shields.io/hexpm/l/fun_with_flags.svg)](https://github.com/tompave/fun_with_flags/blob/master/LICENSE.txt)\n[![ElixirWeekly](https://img.shields.io/badge/featured-ElixirWeekly-8e5ab5.svg)](https://elixirweekly.net/issues/43)\n[![ElixirCasts](https://img.shields.io/badge/featured-ElixirCasts-ff931e.svg)](https://elixircasts.io/feature-flags)\n\nFunWithFlags, the Elixir feature flag library.\n\nIf you're reading this on the [GitHub repo](https://github.com/tompave/fun_with_flags), keep in mind that this readme refers to the `master` branch. For the latest version released on Hex, please check [the readme published with the docs](https://hexdocs.pm/fun_with_flags/readme.html).\n\n---\n\nFunWithFlags is an OTP application that provides a 2-level storage to save and retrieve feature flags, an Elixir API to toggle and query them, and a [web dashboard](#web-dashboard) as control panel.\n\nIt stores flag information in Redis or a relational DB (PostgreSQL, MySQL, or SQLite - with Ecto) for persistence and synchronization across different nodes, but it also maintains a local cache in an ETS table for fast lookups. When flags are added or toggled on a node, the other nodes are notified via PubSub and reload their local ETS caches.","ref":"readme.html"},{"type":"extras","title":"Content - FunWithFlags","doc":"* [What's a Feature Flag](#whats-a-feature-flag)\n* [Usage](#usage)\n  - [Gate Priority and Interactions](#gate-priority-and-interactions)\n  - [Boolean Gate](#boolean-gate)\n  - [Actor Gate](#actor-gate)\n  - [Group Gate](#group-gate)\n  - [Percentage of Time Gate](#percentage-of-time-gate)\n  - [Percentage of Actors Gate](#percentage-of-actors-gate)\n  - [Clearing a Feature Flag's Rules](#clearing-a-feature-flags-rules)\n* [Web Dashboard](#web-dashboard)\n* [Origin](#origin)\n* [So, caching, huh?](#so-caching-huh)\n* [To Do](#to-do)\n* [Installation](#installation)\n* [Configuration](#configuration)\n  - [Persistence Adapters](#persistence-adapters)\n    - [Ecto Multi-tenancy](#ecto-multi-tenancy)\n    - [Ecto Custom Primary Key Types](#ecto-custom-primary-key-types)\n  - [PubSub Adapters](#pubsub-adapters)\n* [Extensibility](#extensibility)\n  - [Custom Persistence Adapters](#custom-persistence-adapters)\n* [Application Start Behaviour](#application-start-behaviour)\n* [Testing](#testing)\n* [Development](#development)\n  - [Working with PubSub Locally](#working-with-pubsub-locally)","ref":"readme.html#content"},{"type":"extras","title":"What's a Feature Flag? - FunWithFlags","doc":"Feature flags, or feature toggles, are boolean values associated to a name. They should be used to control whether some application feature is enabled or disabled, and they are meant to be modified at runtime while an application is running. This is usually done by the people who control the application.\n\nIn their simplest form, flags can be toggled on and off globally. More advanced rules or \"gates\" allow a fine grained control over their status. For example, it's possible to toggle a flag on and off for specific entities or for groups.\n\nThe goal is to have more granular and precise control over what is made available to which users, and when.\nA common use case, in web applications, is to enable a functionality without the need to deploy or restart the server, or to enable it only for internal users to test it before rolling it out to everyone. Another scenario is the ability to quickly disable a functionality if it's causing problems.\nThey can also be used to implement a simple authorization system, for example to an admin area.","ref":"readme.html#what-s-a-feature-flag"},{"type":"extras","title":"Usage - FunWithFlags","doc":"FunWithFlags has a simple API to query and toggle feature flags. Most of the time, you'll call `FunWithFlags.enabled?/2` with the name of the flag and optional arguments.\n\nDifferent kinds of toggle gates are supported:\n\n* **Boolean**: globally on and off.\n* **Actors**: on or off for specific structs or data. The `FunWithFlags.Actor` protocol can be implemented for types and structs that should have specific rules. For example, in web applications it's common to use a `%User{}` struct or equivalent as an actor, or perhaps the current country of the request.\n* **Groups**: on or off for structs or data that belong to a category or satisfy a condition. The `FunWithFlags.Group` protocol can be implemented for types and structs that belong to groups for which a feature flag can be enabled or disabled. For example, one could implement the protocol for a `%User{}` struct to identify administrators.\n* **%-of-Time**: globally on for a percentage of the time. It ignores actors and groups. Mutually exclusive with the %-of-actors gate.\n* **%-of-Actors**: globally on for a percentage of the actors. It only applies when the flag is checked with a specific actor and is ignored when the flag is checked without actor arguments. Mutually exclusive with the %-of-time gate.\n\nBoolean, Actor and Group gates can express either an enabled or disabled state. The percentage gates can only express an enabled state, as disabling something for a percentage of time or actors is logically equivalent to enabling it for the complementary percentage.","ref":"readme.html#usage"},{"type":"extras","title":"Gate Priority and Interactions - FunWithFlags","doc":"The priority order is from most to least specific: `Actors > Groups > Boolean > Percentage`, and it applies to both enabled and disabled gates.\n\nFor example, a disabled group gate takes precedence over an enabled boolean (global) gate for the entities in the group, and a further enabled actor gate overrides the disabled group gate for a specific entity. When an entity belongs to multiple groups with conflicting toggle status, the disabled group gates have precedence over the enabled ones. The percentage gates are checked last, if present, and they're only checked if no other gate is enabled.\n\nAs another example, a flag can have a disabled boolean gate and a 50% enabled %-of-actors gate. When the flag is checked with an actor, it has a (deterministic, consistent and repeatable) 50% chance to be enabled, but when checked without an actor argument it will always be disabled. If we add to the flag a disabled actor gate and an enabled group gate, the flag will be always disabled for the actor, always enabled for any other actor matching the group, have a 50% change to be enabled for any other actor, and always be disabled when checked without actor arguments. If, then, we replace the 50%-of-actors gate with a 50%-of-time gate, the flag will be always disabled for the actor, always enabled for any other actor matching the group, and have a 50% chance to be enabled for any other actor or when checked without an actor argument.","ref":"readme.html#gate-priority-and-interactions"},{"type":"extras","title":"Boolean Gate - FunWithFlags","doc":"The boolean gate is the simplest one. It's either enabled or disabled, globally. It's also the gate with the second lowest priority (it can mask the percentage gates). If a flag is undefined, it defaults to be globally disabled.\n\n```elixir\nFunWithFlags.enabled?(:cool_new_feature)\nfalse\n\n{:ok, true} = FunWithFlags.enable(:cool_new_feature)\n\nFunWithFlags.enabled?(:cool_new_feature)\ntrue\n\n{:ok, false} = FunWithFlags.disable(:cool_new_feature)\n\nFunWithFlags.enabled?(:cool_new_feature)\nfalse\n```","ref":"readme.html#boolean-gate"},{"type":"extras","title":"Actor Gate - FunWithFlags","doc":"This allows you to enable or disable a flag for one or more entities. For example, in web applications it's common to use a `%User{}` struct or equivalent as an actor, or perhaps the data used to represent the current country for an HTTP request. This can be useful to showcase a work-in-progress feature to someone, to gradually rollout a functionality by country, or to dynamically disable some features in some contexts.\n\nActor gates take precedence over the others, both when they're enabled and when they're disabled. They can be considered as toggle overrides.\n\nIn order to be used as an actor, an entity must implement the `FunWithFlags.Actor` protocol. This can be implemented for custom structs or literally any other type.\n\n\n```elixir\ndefmodule MyApp.User do\n  defstruct [:id, :name]\nend\n\ndefimpl FunWithFlags.Actor, for: MyApp.User do\n  def id(%{id: id}) do\n    \"user:#{id}\"\n  end\nend\n\nbruce = %MyApp.User{id: 1, name: \"Bruce\"}\nalfred = %MyApp.User{id: 2, name: \"Alfred\"}\n\nFunWithFlags.Actor.id(bruce)\n\"user:1\"\nFunWithFlags.Actor.id(alfred)\n\"user:2\"\n\ndefimpl FunWithFlags.Actor, for: Map do\n  def id(%{actor_id: actor_id}) do\n    \"map:#{actor_id}\"\n  end\n\n  def id(map) do\n    map\n    |> inspect()\n    |> (&:crypto.hash(:md5, &1)).()\n    |> Base.encode16\n    |> (&\"map:#{&1}\").()\n  end\nend\n\nFunWithFlags.Actor.id(%{actor_id: \"bar\"})\n\"map:bar\"\nFunWithFlags.Actor.id(%{foo: \"bar\"})\n\"map:E0BB5BA6873E3AC34B0B6928190C1F2B\"\n```\n\nWith the protocol implemented, actors can be used with the library functions:\n\n```elixir\n{:ok, true} = FunWithFlags.enable(:restful_nights)\n{:ok, false} = FunWithFlags.disable(:restful_nights, for_actor: bruce)\n{:ok, true} = FunWithFlags.enable(:batmobile, for_actor: bruce)\n\nFunWithFlags.enabled?(:restful_nights)\ntrue\nFunWithFlags.enabled?(:batmobile)\nfalse\n\nFunWithFlags.enabled?(:restful_nights, for: alfred)\ntrue\nFunWithFlags.enabled?(:batmobile, for: alfred)\nfalse\n\nFunWithFlags.enabled?(:restful_nights, for: bruce)\nfalse\nFunWithFlags.enabled?(:batmobile, for: bruce)\ntrue\n```\n\nActor identifiers must be globally unique binaries. Since supporting multiple kinds of actors is a common requirement, all the examples use the common technique of namespacing the IDs:\n\n```elixir\ndefimpl FunWithFlags.Actor, for: MyApp.User do\n  def id(user) do\n    \"user:#{user.id}\"\n  end\nend\n\ndefimpl FunWithFlags.Actor, for: MyApp.Country do\n  def id(country) do\n    \"country:#{country.iso3166}\"\n  end\nend\n```","ref":"readme.html#actor-gate"},{"type":"extras","title":"Group Gate - FunWithFlags","doc":"Group gates are similar to actor gates, but they apply to a category of entities rather than specific ones. They can be toggled on or off for the _name of the group_ instead of a specific term.\n\nGroup gates take precedence over boolean gates but are overridden by actor gates.\n\nGroup names can be binaries or atoms. Atoms are supported for retro-compatibility with versions `<= 0.9` and binaries are therefore preferred. In fact, atoms are internally converted to binaries and are then stored and later retrieved as binaries.\n\nThe semantics to determine which entities belong to which groups are application specific.\nEntities could have an explicit list of groups they belong to, or the groups could be abstract and inferred from some other attribute. For example, an `:employee` group could comprise all `%User{}` structs with an email address matching the company domain, or an `:admin` group could be made of all users with `%User{admin: true}`.\n\nIn order to be affected by a group gate, an entity should implement the `FunWithFlags.Group` protocol. The protocol automatically falls back to a default `Any` implementation, which states that any entity belongs to no group at all. This makes it possible to safely use \"normal\" actors when querying group gates, and to implement the protocol only for structs and types for which it matters.\n\nThe protocol can be implemented for custom structs or literally any other type.\n\n\n```elixir\ndefmodule MyApp.User do\n  defstruct [:email, admin: false, groups: []]\nend\n\ndefimpl FunWithFlags.Group, for: MyApp.User do\n  def in?(%{email: email}, \"employee\"), do: Regex.match?(~r/@mycompany.com$/, email)\n  def in?(%{admin: is_admin}, \"admin\"), do: !!is_admin\n  def in?(%{groups: list}, group_name), do: group_name in list\nend\n\nelisabeth = %MyApp.User{email: \"elisabeth@mycompany.com\", admin: true, groups: [\"engineering\", \"product\"]}\nFunWithFlags.Group.in?(elisabeth, \"employee\")\ntrue\nFunWithFlags.Group.in?(elisabeth, \"admin\")\ntrue\nFunWithFlags.Group.in?(elisabeth, \"engineering\")\ntrue\nFunWithFlags.Group.in?(elisabeth, \"marketing\")\nfalse\n\ndefimpl FunWithFlags.Group, for: Map do\n  def in?(%{group: group_name}, group_name), do: true\n  def in?(_, _), do: false\nend\n\nFunWithFlags.Group.in?(%{group: \"dumb_tests\"}, \"dumb_tests\")\ntrue\n```\n\nWith the protocol implemented, actors can be used with the library functions:\n\n```elixir\nFunWithFlags.disable(:database_access)\nFunWithFlags.enable(:database_access, for_group: \"engineering\")\n\nFunWithFlags.enabled?(:database_access)\nfalse\nFunWithFlags.enabled?(:database_access, for: elisabeth)\ntrue\n```","ref":"readme.html#group-gate"},{"type":"extras","title":"Percentage of Time Gate - FunWithFlags","doc":"%-of-time gates are similar to boolean gates, but they allow to enable a flag for a percentage of the time. In practical terms, this means that a percentage of the `enabled?()` calls for a flag will return true, regardless of the presence of an actor argument.\n\nWhen a %-of-time gate is checked a [pseudo-random number is generated](http://erlang.org/doc/man/rand.html#uniform-1) and compared with the percentage value of the gate. If the result of the random roll is lower than the gate's percentage value, the gate is considered enabled. So, at the risk of stating the obvious and for the sake of clarity, a 90% gate is enabled more often than a 10% gate.\n\n%-of-time gates are useful to gradually introduce alternative code paths that either have the same effects of the old ones, or don't have effects visible to the users. This last point is important, because with a %-of-time gate the application will behave differently on a pseudo-random basis.\n\nA good use case for %-of-time gates is to safely test the correctness or performance and load characteristics of an alternative implementation of a functionality.\n\nFor example:\n\n```elixir\nFunWithFlags.clear(:alternative_implementation)\nFunWithFlags.enable(:alternative_implementation, for_percentage_of: {:time, 0.05})\n\ndef foo(bar) do\n  if FunWithFlags.enabled?(:alternative_implementation) do\n    new_foo(bar)\n  else\n    old_foo(bar)\n  end\nend\n```\n\nThe %-of-time gate is incompatible and mutually exclusive with the %-of-actors gate, and it replaces it when it gets set. While there are ways to make them work together, it would needlessly overcomplicate the priority rules.","ref":"readme.html#percentage-of-time-gate"},{"type":"extras","title":"Percentage of Actors Gate - FunWithFlags","doc":"%-of-actors gates are similar to the %-of-time gates, but instead of using a pseudo-random chance they calculate the actor scores using a deterministic, consistent and repeatable function that factors in the flag name. At a high level:\n\n```elixir\nactor\n|> FunWithFlags.Actor.id()\n|> sha256_hash(flag_name)\n|> hash_to_percentage()\n```\n\nSince the scores depend on both the actor ID and the flag name, they're guaranteed to always be the same for each actor-flag combination. At the same time, the same actor will have different scores for different flags, and each flag will have a uniform distribution of scores for all the actors.\n\nJust like for the %-of-time gates, an actor's score is compared with the gate's percentage value and, if lower, the gate will result enabled.\n\nA practical example, based on the `FunWithFlags.Actor` protocol set up from the previous sections:\n\n```elixir\ndefmodule MyApp.User do\n  defstruct [:id, :name]\nend\n\ndefimpl FunWithFlags.Actor, for: MyApp.User do\n  def id(%{id: id}) do\n    \"user:#{id}\"\n  end\nend\n\nfrodo  = %MyApp.User{id: 1, name: \"Frodo Baggins\"}\nsam    = %MyApp.User{id: 2, name: \"Samwise Gamgee\"}\npippin = %MyApp.User{id: 3, name: \"Peregrin Took\"}\nmerry  = %MyApp.User{id: 4, name: \"Meriadoc Brandybuck\"}\n\nFunWithFlags.Actor.Percentage.score(frodo, :pipeweed)\n0.8658294677734375\nFunWithFlags.Actor.Percentage.score(sam, :pipeweed)\n0.68426513671875\nFunWithFlags.Actor.Percentage.score(pippin, :pipeweed)\n0.510528564453125\nFunWithFlags.Actor.Percentage.score(merry, :pipeweed)\n0.2617645263671875\n\n{:ok, true} = FunWithFlags.enable(:pipeweed, for_percentage_of: {:actors, 0.60})\n\nFunWithFlags.enabled?(:pipeweed, for: frodo)\nfalse\nFunWithFlags.enabled?(:pipeweed, for: sam)\nfalse\nFunWithFlags.enabled?(:pipeweed, for: pippin)\ntrue\nFunWithFlags.enabled?(:pipeweed, for: merry)\ntrue\n\n{:ok, true} = FunWithFlags.enable(:pipeweed, for_percentage_of: {:actors, 0.685})\n\nFunWithFlags.enabled?(:pipeweed, for: sam)\ntrue\n\nFunWithFlags.Actor.Percentage.score(pippin, :pipeweed)\n0.510528564453125\nFunWithFlags.Actor.Percentage.score(pippin, :mushrooms)\n0.6050872802734375\nFunWithFlags.Actor.Percentage.score(pippin, :palantir)\n0.144073486328125\n```\n\n\nOnce a %-of-actors gate has been defined for a flag, the same actor will always see the same result (unless its actor or group gates are set, or the flag gets globally enabled). Also, this means that as long the percentage value of the gate will increase and never decrease, actors for which the gate has been enabled will always see it enabled.\n\nThis is ideal to gradually roll out new functionality to users.\n\nFor example, in a Phoenix application:\n\n```elixir\nFunWithFlags.clear(:new_design)\nFunWithFlags.enable(:new_design, for_percentage_of: {:actors, 0.2})\nFunWithFlags.enable(:new_design, for_group: \"beta_testers\")\n\n\ndefmodule MyPhoenixApp.MyView do\n  use MyPhoenixApp, :view\n\n  def render(\"my_template.html\", assigns) do\n    if FunWithFlags.enabled?(:new_design, for: assigns.user) do\n      render(\"new_template.html\", assigns)\n    else\n      render(\"old_template.html\", assigns)\n    end\n  end\nend\n```\n\nThe %-of-actors gate is incompatible and mutually exclusive with the %-of-time gate, and it replaces it when it gets set. While there are ways to make them work together, it would needlessly overcomplicate the priority rules.","ref":"readme.html#percentage-of-actors-gate"},{"type":"extras","title":"Clearing a Feature Flag's Rules - FunWithFlags","doc":"Sometimes enabling or disabling a gate is not what you want, and removing that gate's rules would be more correct. For example, if you don't need anymore to explicitly enable or disable a flag for an actor or for a group, and the default state should be used instead, clearing the gate is the right choice.\n\nMore examples:\n\n```elixir\nalias FunWithFlags.TestUser, as: User\nharry = %User{id: 1, name: \"Harry Potter\", groups: [\"wizards\", \"gryffindor\"]}\nhagrid = %User{id: 2, name: \"Rubeus Hagrid\", groups: [\"wizards\", \"gamekeeper\"]}\ndudley = %User{id: 3, name: \"Dudley Dursley\", groups: [\"muggles\"]}\nFunWithFlags.disable(:wands)\nFunWithFlags.enable(:wands, for_group: \"wizards\")\nFunWithFlags.disable(:wands, for_actor: hagrid)\n\nFunWithFlags.enabled?(:wands)\nfalse\nFunWithFlags.enabled?(:wands, for: harry)\ntrue\nFunWithFlags.enabled?(:wands, for: hagrid)\nfalse\nFunWithFlags.enabled?(:wands, for: dudley)\nfalse\n\nFunWithFlags.clear(:wands, for_actor: hagrid)\n\nFunWithFlags.enabled?(:wands, for: hagrid)\ntrue\n\nFunWithFlags.clear(:wands, for_group: \"wizards\")\n\nFunWithFlags.enabled?(:wands, for: hagrid)\nfalse\nFunWithFlags.enabled?(:wands, for: harry)\nfalse\n\nFunWithFlags.enable(:magic_powers, for_percentage_of: {:time, 0.0001})\nFunWithFlags.clear(:magic_powers, for_percentage: true)\n```\n\nFor completeness, clearing the boolean gate is also supported.\n\n```elixir\nFunWithFlags.enable(:wands)\n\nFunWithFlags.enabled?(:wands)\ntrue\nFunWithFlags.enabled?(:wands, for: harry)\ntrue\nFunWithFlags.enabled?(:wands, for: hagrid)\nfalse\nFunWithFlags.enabled?(:wands, for: dudley)\ntrue\n\nFunWithFlags.clear(:wands, boolean: true)\n\nFunWithFlags.enabled?(:wands)\nfalse\nFunWithFlags.enabled?(:wands, for: harry)\ntrue\nFunWithFlags.enabled?(:wands, for: hagrid)\nfalse\nFunWithFlags.enabled?(:wands, for: dudley)\nfalse\n```\n\nIt's also possible to clear an entire flag.\n\n```elixir\nFunWithFlags.clear(:wands)\n\nFunWithFlags.enabled?(:wands)\nfalse\nFunWithFlags.enabled?(:wands, for: harry)\nfalse\nFunWithFlags.enabled?(:wands, for: hagrid)\nfalse\nFunWithFlags.enabled?(:wands, for: dudley)\nfalse\n```","ref":"readme.html#clearing-a-feature-flag-s-rules"},{"type":"extras","title":"Web Dashboard - FunWithFlags","doc":"An optional extension of this library is [`FunWithFlags.UI`](https://github.com/tompave/fun_with_flags_ui), a web graphical control panel. It's a Plug, so it can be embedded in a host Phoenix or Plug application or served standalone.","ref":"readme.html#web-dashboard"},{"type":"extras","title":"Origin - FunWithFlags","doc":"This library is heavily inspired by the [flipper Ruby gem](https://github.com/jnunemaker/flipper).\n\nHaving used Flipper in production at scale, this project aims to improve in two main areas:\n\n* Minimize the load on the persistence layer: feature flags are not toggled _that_ often, and there is no need to query Redis or the DB for each check.\n* Be more reliable: it should keep working with the latest cached values even if Redis becomes unavailable, although with the risk of nodes getting out of sync. (if the DB becomes unavailable, feature flags are probably the last of your problems)\n\nJust as Elixir and Phoenix are meant to scale better than Ruby on Rails with high levels of traffic and concurrency, FunWithFlags should aim to be more scalable and reliable than Flipper.","ref":"readme.html#origin"},{"type":"extras","title":"So, caching, huh? - FunWithFlags","doc":"> There are only two hard things in Computer Science: cache invalidation and naming things.\n>\n> -- Phil Karlton\n\nThe reason to add an ETS cache is that, most of the time, feature flags can be considered static values. Doing a round-trip to the DB (Redis, PostgreSQL or MySQL) is expensive in terms of time and in terms of resources, especially if multiple flags must be checked during a single web request. In the worst cases, the load on the DB can become a cause of concern, a performance bottleneck or the source of a system failure.\n\nOften the solution is to memoize the flag values _in the context of the web request_, but the approach can be extended to the scope of the entire server. This is what FunWithFlags does, as each application node/instance caches the flags in an ETS table.\n\nOf course, caching adds a different kind of complexity and there are some pros and cons. When a flag is created or updated the ETS cache on the local node is updated immediately, and the main problem is synchronizing the flag data across the other application nodes that should share the same view of the world.\n\nFor example, if we have two or more nodes running the application, and on one of them an admin user updates a flag that the others have already cached, or creates a flag that the others have already looked up (and cached as \"disabled\"), then the other nodes must  be notified of the changes.\n\nFunWithFlags uses three mechanisms to deal with the problem:\n\n1. Use PubSub to emit change notifications. All nodes subscribe to the same channel and reload flags in the ETS cache when required.\n2. If that fails, the cache has a configurable TTL. Reading from the DB every few minutes is still better than doing so 30k times per second.\n3. If that doesn't work, it's possible to disable the cache and just read from the DB all the time. That's what Flipper does.\n\nIn terms of performance, very synthetic benchmarks (where the DBs run on the same machine as the Beam code, so with no network hop but sharing the CPU) show that the ETS cache makes querying the FunWithFlags interface between 10 and 20 times faster than going directly to Redis, and between 20 and 40 times faster than going directly to Postgres. The variance depends on the complexity of the flag data to be retrieved.","ref":"readme.html#so-caching-huh"},{"type":"extras","title":"To Do - FunWithFlags","doc":"* Add some optional randomness to the TTL, so that Redis or the DB don't get hammered at constant intervals after a server restart.","ref":"readme.html#to-do"},{"type":"extras","title":"Installation - FunWithFlags","doc":"The package can be installed by adding `fun_with_flags` to your list of dependencies in `mix.exs`.\n\nIn order to have a small installation footprint, the dependencies for the different adapters are all optional. You must explicitly require the ones you wish to use.\n\n```elixir\ndef deps do\n  [\n    {:fun_with_flags, \"~> 1.12.0\"},\n\n    # either:\n    {:redix, \"~> 0.9\"},\n    # or:\n    {:ecto_sql, \"~> 3.0\"},\n\n    # optionally, if you don't want to use Redis' builtin pubsub\n    {:phoenix_pubsub, \"~> 2.0\"},\n  ]\nend\n```\n\nUsing `ecto_sql` for persisting the flags also requires an ecto adapter, e.g. `postgrex`, `mariaex` or `myxql`. Please refer to the Ecto documentation for the details.\n\nSince FunWithFlags depends on an Elixir more recent than 1.4, there is [no need to explicitly declare the application](https://github.com/elixir-lang/elixir/blob/v1.4/CHANGELOG.md#application-inference).\n\nIf you need to customize how the `:fun_with_flags` application is loaded and started, refer to the [Application Start Behaviour](#application-start-behaviour) section, below in this document.","ref":"readme.html#installation"},{"type":"extras","title":"Configuration - FunWithFlags","doc":"The library can be configured in host applications through Mix and the `config.exs` file. This example shows some default values:\n\n```elixir\nconfig :fun_with_flags, :cache,\n  enabled: true,\n  ttl: 900 # in seconds\n\n# the Redis persistence adapter is the default, no need to set this.\nconfig :fun_with_flags, :persistence,\n  [adapter: FunWithFlags.Store.Persistent.Redis]\n\n# this can be disabled if you are running on a single node and don't need to\n# sync different ETS caches. It won't have any effect if the cache is disabled.\n# The Redis PuSub adapter is the default, no need to set this.\nconfig :fun_with_flags, :cache_bust_notifications,\n  [enabled: true, adapter: FunWithFlags.Notifications.Redis]\n\n# Notifications can also be disabled, which will also remove the Redis/Redix dependency\nconfig :fun_with_flags, :cache_bust_notifications, [enabled: false]\n```\n\nWhen using Redis for persistence and/or cache-busting PubSub it is necessary to configure the connection to the Redis instance. These options can be omitted if Redis is not being used. For example, the defaults:\n\n```elixir\n# the Redis options will be forwarded to Redix.\nconfig :fun_with_flags, :redis,\n  host: \"localhost\",\n  port: 6379,\n  database: 0\n\n# a URL string can be used instead\nconfig :fun_with_flags, :redis, \"redis://localhost:6379/0\"\n\n# or a {URL, [opts]} tuple\nconfig :fun_with_flags, :redis, {\"redis://localhost:6379/0\", socket_opts: [:inet6]}\n\n# a {:system, name} tuple can be used to read from the environment\nconfig :fun_with_flags, :redis, {:system, \"REDIS_URL\"}\n```\n\n[Redis Sentinel](https://redis.io/docs/manual/sentinel/) is also supported. See the [Redix docs](https://github.com/whatyouhide/redix/tree/v1.1.5#redis-sentinel) for more details.\n\n```elixir\nconfig :fun_with_flags, :redis,\n  sentinel: [\n    sentinels: [\"redis:://locahost:1234/1\"],\n    group: \"primary\",\n  ],\n  database: 5\n```","ref":"readme.html#configuration"},{"type":"extras","title":"Persistence Adapters - FunWithFlags","doc":"The library comes with two persistence adapters for the [`Redix`](https://hex.pm/packages/redix) and [`Ecto`](https://hex.pm/packages/ecto) libraries, that allow to persist feature flag data in Redis, PostgreSQL, MySQL, or SQLite. In order to use any of them, you must declare the correct optional dependency in the Mixfile (see the [installation](#installation) instructions, above).\n\nThe Redis adapter is the default and there is no need to explicitly declare it. All it needs is the Redis connection configuration.\n\nIn order to use the Ecto adapter, an Ecto repo must be provided in the configuration. FunWithFlags expects the Ecto repo to be initialized by the host application, which also needs to start and supervise any required processes. If using Phoenix this is managed automatically by the framework, and it's fine to use the same repo used by the rest of the application.\n\nOnly PostgreSQL (via [`postgrex`](https://hex.pm/packages/postgrex)), MySQL (via [`mariaex`](https://hex.pm/packages/mariaex) or [`myxql`](https://hex.pm/packages/myxql)), and SQLite (via [`ecto_sqlite3`](https://hex.pm/packages/ecto_sqlite3)) are supported at the moment. Support for other RDBMSs might come in the future.\n\nTo configure the Ecto adapter:\n\n```elixir\n# Normal Phoenix and Ecto configuration.\n# The repo can either use the Postgres or MySQL adapter.\nconfig :my_app, ecto_repos: [MyApp.Repo]\nconfig :my_app, MyApp.Repo,\n  username: \"my_db_user\",\n  password: \"my secret db password\",\n  database: \"my_app_dev\",\n  hostname: \"localhost\",\n  pool_size: 10\n\n# FunWithFlags configuration.\nconfig :fun_with_flags, :persistence,\n  adapter: FunWithFlags.Store.Persistent.Ecto,\n  repo: MyApp.Repo,\n  ecto_table_name: \"your_table_name\", # optional, defaults to \"fun_with_flags_toggles\"\n  ecto_primary_key_type: :binary_id # optional, defaults to :id\n  # For the primary key type, see also: https://hexdocs.pm/ecto/3.10.3/Ecto.Schema.html#module-schema-attributes\n```\n\nIt's also necessary to create the DB table that will hold the feature flag data. To do that, [create a new migration](https://hexdocs.pm/ecto_sql/Mix.Tasks.Ecto.Gen.Migration.html) in your project and copy the contents of [the provided migration file](https://github.com/tompave/fun_with_flags/blob/master/priv/ecto_repo/migrations/00000000000000_create_feature_flags_table.exs). Then [run the migration](https://hexdocs.pm/ecto_sql/Mix.Tasks.Ecto.Migrate.html).\n\nWhen using the Ecto persistence adapter, FunWithFlags will annotate all queries using the [Ecto Repo Query API](https://hexdocs.pm/ecto/3.8.4/Ecto.Repo.html#query-api) with a custom option: `[fun_with_flags: true]`. This is done to make it easier to identify FunWithFlags queries when working with Ecto customization hooks, e.g. the [`Ecto.Repo.prepare_query/3` callback](https://hexdocs.pm/ecto/3.8.4/Ecto.Repo.html#c:prepare_query/3). Since this sort of annotations via custom query options are only useful with the Ecto Query API ([context](https://elixirforum.com/t/proposal-prepare-query-for-write-operations/50510)), other repo functions are not annotated with the custom option.\n\n#### Ecto Multi-tenancy\n\nIf you followed the Ecto guide on setting up [multi-tenancy with foreign keys](https://hexdocs.pm/ecto/3.8.4/multi-tenancy-with-foreign-keys.html), you must add an exception for queries originating from FunWithFlags. As mentioned in the section above, these queries have a custom query option named `:fun_with_flags` set to `true`:\n\n```elixir\n# Sample code, only relevant if you followed the Ecto guide on multi tenancy with foreign keys.\ndefmodule MyApp.Repo do\n  use Ecto.Repo, otp_app: :my_app\n\n  require Ecto.Query\n\n  @impl true\n  def prepare_query(_operation, query, opts) do\n    cond do\n      # add the check for opts[:fun_with_flags] here:\n      opts[:skip_org_id] || opts[:schema_migration] || opts[:fun_with_flags] ->\n        {query, opts}\n\n      org_id = opts[:org_id] ->\n        {Ecto.Query.where(query, org_id: ^org_id), opts}\n\n      true ->\n        raise \"expected org_id or skip_org_id to be set\"\n    end\n  end\nend\n```\n\n#### Ecto Custom Primary Key Types\n\nThe library defaults to using an integer (`bigserial`) as the type of the `id` primary key column. If, for any reason, you need the ID to be a UUID, you can configure it to be of type `:binary_id`. To do that, you need to:\n\n  1. Set the `:ecto_primary_key_type` configuration option to `:binary_id`.\n  2. Use `:binary_id` as the type of the `:id` column in the [provided migration file](https://github.com/tompave/fun_with_flags/blob/master/priv/ecto_repo/migrations/00000000000000_create_feature_flags_table.exs).","ref":"readme.html#persistence-adapters"},{"type":"extras","title":"PubSub Adapters - FunWithFlags","doc":"The library comes with two PubSub adapters for the [`Redix`](https://hex.pm/packages/redix) and [`Phoenix.PubSub`](https://hex.pm/packages/phoenix_pubsub) libraries. In order to use any of them, you must declare the correct optional dependency in the Mixfile. (see the [installation](#installation) instructions, below)\n\nThe Redis PubSub adapter is the default and doesn't need to be explicitly configured. It can only be used in conjunction with the Redis persistence adapter however, and is not available when using Ecto for persistence. When used, it connects directly to the Redis instance used for persisting the flag data.\n\nThe Phoenix PubSub adapter uses the high level API of `Phoenix.PubSub`, which means that under the hood it could use either its PG2 or Redis adapters, and this library doesn't need to know. It's provided as a convenient way to leverage distributed Erlang when using FunWithFlags in a Phoenix application, although it can be used independently (without the rest of the Phoenix framework) to add PubSub to Elixir apps running on Erlang clusters.  \nFunWithFlags expects the `Phoenix.PubSub` process to be started by the host application, and in order to use this adapter the client (name or PID) must be provided in the configuration.\n\nFor example, in Phoenix (>= 1.5.0) it would be:\n\n```elixir\n# normal Phoenix configuration\nconfig :my_app, MyApp.Web.Endpoint,\n  pubsub_server: MyApp.PubSub\n\n# FunWithFlags configuration\nconfig :fun_with_flags, :cache_bust_notifications,\n  enabled: true,\n  adapter: FunWithFlags.Notifications.PhoenixPubSub,\n  client: MyApp.PubSub\n```\n\nOr, without Phoenix:\n\n```elixir\n# possibly in the application's supervision tree\nchildren = [\n  {Phoenix.PubSub, [name: :my_pubsub_process_name, adapter: Phoenix.PubSub.PG2]}\n]\nopts = [strategy: :one_for_one, name: MyApp.Supervisor]\n{:ok, _pid} = Supervisor.start_link(children, opts)\n\n# config/config.exs\nconfig :fun_with_flags, :cache_bust_notifications,\n  enabled: true,\n  adapter: FunWithFlags.Notifications.PhoenixPubSub,\n  client: :my_pubsub_process_name\n```","ref":"readme.html#pubsub-adapters"},{"type":"extras","title":"Extensibility - FunWithFlags","doc":"","ref":"readme.html#extensibility"},{"type":"extras","title":"Custom Persistence Adapters - FunWithFlags","doc":"This library aims to be extensible and allows users to provide their own persistence layer.\n\nThis is supported through [`FunWithFlags.Store.Persistent`](https://github.com/tompave/fun_with_flags/blob/master/lib/fun_with_flags/store/persistent.ex), a generic persistence [behaviour](https://hexdocs.pm/elixir/typespecs.html#behaviours) that is adopted by the builtin Redis and Ecto adapters.\n\nCustom persistence adapters can adopt the behaviour and then be configured as the persistence module in the Mix config of the user applications.\n\nFor example, an application can define this module:\n\n```elixir\ndefmodule MyApp.MyAlternativeFlagStore do\n  @behaviour FunWithFlags.Store.Persistent\n  # implement all the behaviour's callback\nend\n```\n\nAnd then configure the library to use it:\n\n```elixir\nconfig :fun_with_flags, :persistence, adapter: MyApp.MyAlternativeFlagStore\n```","ref":"readme.html#custom-persistence-adapters"},{"type":"extras","title":"Application Start Behaviour - FunWithFlags","doc":"As explained in the [Installation](#installation) section, above in this document, the `:fun_with_flags` application will start automatically when you add the package as a dependency in your Mixfile. The `:fun_with_flags` application starts its own supervision tree which manages all required processes and is provided by the `FunWithFlags.Supervisor` module.\n\nSometimes, this can cause issues and race conditions if FunWithFlags is configured to rely on Erlang processes that are owned by another application. For example, if you have configured the `Phoenix.PubSub` cache-busting notification adapter, one of FunWithFlag's processes will immediately try to subscribe to its notifications channel using the provided PubSub process identifier. If that process is not available, FunWithFlags will retry a few times and then give up and raise an exception. This will become a problem if you're using FunWithFlags in a large application (e.g. a Phoenix app) and the `:fun_with_flags` application starts much faster than the Phoenix supervision tree.\n\nIn these cases, it's better to directly control how FunWithFlags starts its processes.\n\nThe first step is to add the `FunWithFlags.Supervisor` module directly to the supervision tree of the host application. For example, in a Phoenix app it would look like this:\n\n```diff\ndefmodule MyPhoenixApp.Application do\n  @moduledoc false\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      MyPhoenixApp.Repo,\n      MyPhoenixAppWeb.Telemetry,\n      {Phoenix.PubSub, name: MyPhoenixApp.PubSub},\n      MyPhoenixAppWeb.Endpoint,\n+     FunWithFlags.Supervisor,\n    ]\n\n    opts = [strategy: :one_for_one, name: MyPhoenixApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  # ...\n```\n\nThen it's necessary to configure the Mix project to not start the `:fun_with_flags` application automatically. This can be accomplished in the Mixfile in a number of ways, for example: (**Note**: These are alternative solutions, you don't need to do both. You must decide which is more appropriate for your setup.)\n\n* **Option A**: Declare the `:fun_with_flags` dependency with either the `runtime: false` or `app: false` options. ([docs](https://hexdocs.pm/mix/1.11.3/Mix.Tasks.Deps.html#module-dependency-definition-options))\n\n```diff\n- {:fun_with_flags, \"~> 1.6\"},\n+ {:fun_with_flags, \"~> 1.6\", runtime: false},\n```\n\nIf you use releases then you'll also need to modify the `releases` section in `mix.exs` so that it loads the `fun_with_flags` application explicitly (since `runtime: false` / `app: false` will exclude it from the assembled release).\n\n```diff\ndef project do\n  [\n    app: :my_phoenix_app,\n+   releases: [\n+     my_phoenix_app: [\n+       applications: [\n+         fun_with_flags: :load\n+       ]\n+    ]\n  ]\nend\n```\n\n* **Option B**: Declare that the `:fun_with_flags` application is managed directly by your host application ([docs](https://hexdocs.pm/mix/1.11.3/Mix.Tasks.Compile.App.html)).\n\n```diff\n  def application do\n    [\n      mod: {MyPhoenixApp.Application, []},\n+     included_applications: [:fun_with_flags],\n      extra_applications: [:logger, :runtime_tools]\n    ]\n  end\n```\n\nThe result of those changes is that the `:fun_with_flags` application won't be loaded and started automatically, and therefore the FunWithFlags supervision tree won't risk to be started before the other processes in the host Phoenix application. Rather, the supervision tree will start alongside the other core Phoenix processes.\n\nOne final note on this topic is that if you're also using [`FunWithFlags.UI`](https://github.com/tompave/fun_with_flags_ui) (refer to the [Web Dashboard](#web-dashboard) section, above in this document), then that will need to be configured as well. The reason is that `:fun_with_flags` is a dependency of `:fun_with_flags_ui`, so including the latter as a dependency will cause the former to be auto-started despite the configuration described above. To avoid this, the same configuration should be used for the `:fun_with_flags_ui` dependency, regardless of the approach used (Option A: `runtime: false`, `app: false`; or Option B: `included_applications`).","ref":"readme.html#application-start-behaviour"},{"type":"extras","title":"Testing - FunWithFlags","doc":"This library depends on Redis, PostgreSQL and MySQL, and you'll need them installed and running on your system in order to run the complete test suite. The tests will use the [Redis db number 5](https://github.com/tompave/fun_with_flags/blob/master/test/support/test_utils.ex#L4) and then clean after themselves, but it's safer to start Redis in a directory where there is no `dump.rdb` file you care about to avoid issues. The Ecto tests will use the SQL sandbox and all transactions will be automatically rolled back.\n\nTo setup the test DB for the Ecto persistence tests, run:\n\n```\nMIX_ENV=test PERSISTENCE=ecto mix do ecto.create, ecto.migrate              # for postgres\nrm -rf _build/test/lib/fun_with_flags/\nMIX_ENV=test PERSISTENCE=ecto RDBMS=mysql mix do ecto.create, ecto.migrate  # for mysql\nrm -rf _build/test/lib/fun_with_flags/\nMIX_ENV=test PERSISTENCE=ecto RDBMS=sqlite mix do ecto.create, ecto.migrate  # for sqlite\n\n```\n\nThen, to run all the tests:\n```\n$ mix test.all\n```\n\nThe `test.all` task will run the test suite multiple times with different configurations to exercise a matrix of options and adapters.\n\nThe Mixfile defines a few other helper tasks that allow to run the test suite with some more specific configurations.","ref":"readme.html#testing"},{"type":"extras","title":"Development - FunWithFlags","doc":"Like for testing, developing FunWithFlags requires local installations of Redis, PostgreSQL and MySQL. For work that doesn't touch the persistence adapters too closely, it's possibly simpler to just run FunWithFlags with Redis and then let CI run the tests with the other adapters.\n\nA common workflow is to run the tests and interact with the package API in `iex`.\n\nWith the [default configuration](https://github.com/tompave/fun_with_flags/blob/master/config/config.exs), `iex -S mix` will compile and load FunWithFlags with Redis persistence and Redis PubSub. To compile and run the package in `iex` with Ecto and Phoenix PubSub support instead, use these commands:\n\n```shell\nbin/console_ecto postgres\nbin/console_ecto mysql\n```\n\nThis package uses the [credo](https://hex.pm/packages/credo) and [dialyxir](https://hex.pm/packages/dialyxir) ([dialyzer](https://erlang.org/doc/man/dialyzer.html)) packages to help with local development. Their mix tasks can be executed in the root directory of the project:\n\n```shell\nmix credo\nmix dialyzer\n```","ref":"readme.html#development"},{"type":"extras","title":"Working with PubSub Locally - FunWithFlags","doc":"It's possible to test the PubSub functionality locally, in `iex`.\n\nWhen using Redis, it's enough to start two `iex -S mix` sessions in two terminals, and they'll talk with one another via Redis.\n\nWhen using `Phoenix.PubSub` (which is typically the case with `Ecto`), then the process is similar but you must establish a connection between the two Erlang nodes running in the two terminals. There are a number of ways to do this, and the simplest is to do it manually within `iex`.\n\nSteps:\n\n1. Run `bin/console_pubsub foo` in one terminal.\n2. Run `bin/console_pubsub bar` in another terminal.\n3. In either terminal, grab the current name with `Node.self()`. (The name will also be shown in the `iex` prompts).\n4. In the other terminal, run `Node.connect(:\"THE_OTHER_NODE_NAME\")`. Keep in mind that the names are atoms.\n5. In either terminal, run `Node.list()` to check that there is a connection.\n\nDone that, modifying any flag data in either terminal will notify the other one via PubSub.","ref":"readme.html#working-with-pubsub-locally"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v1.12.0 - Changelog","doc":"* Add support for Elixir 1.17 and 1.16. Drop support for Elixir 1.13 and 1.14. Elixir >= 1.15 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n    * Note: The FunWithFlags package has usually increased its supported Elixir version one at a time, usually supporting the last three minor versions. With this version `v1.12.0` it fast-forwards to Elixir 1.17 from 1.15, and drops support for Elixir versions 1.13 and 1.14 at the same time. That's because a new version of the package wasn't released to support Elixir 1.16 specifically, when it came out, and now there are two new Elixir versions out there. The good news is that this is just a formality to be clear on which Elixir and OTP versions are officially supported, and the FunWithFlags package actually has been working pretty well with new Elixir versions, and even with older Elixir versions that are formally unsupported.\n* Fix incorrect return value in `FunWithFlags.disable/2` typespec. (Thanks [mbuffa](https://github.com/mbuffa), [pull/170](https://github.com/tompave/fun_with_flags/pull/170))\n* Improved the performance of checking a flag with many actor or group gates. (Thanks [up2jj](https://github.com/up2jj), [pull/172](https://github.com/tompave/fun_with_flags/pull/172))","ref":"changelog.html#v1-12-0"},{"type":"extras","title":"v1.11.0 - Changelog","doc":"* Add support for Elixir 1.15. Drop support for Elixir 1.12. Elixir >= 1.13 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* Drop support for Erlang/OTP 23, and Erlang/OTP >= 24 is now required. Dropping support for older versions of Erlang/OTP simply means that this package is not tested with them in CI, and that no compatibility issues are considered bugs.\n* Remove from the repo the [DB migration](https://github.com/tompave/fun_with_flags/blob/v1.1.0/priv/ecto_repo/migrations/00000000000001_ensure_columns_are_not_null.exs) added in [v1.1.0](https://github.com/tompave/fun_with_flags/blob/master/CHANGELOG.md#v110) (November 2018), as an upgrade step. After almost 5 years, chances are that users of the library are already using the correct schema, and that extra \"upgrade\" migration is incompatible with RDBMS other than Postgres and MySQL.\n* Add support for SQLite with the [`ecto_sqlite3`](https://hex.pm/packages/ecto_sqlite3) adapter. (Thanks [tylerbarker](https://github.com/tylerbarker), [pull/151](https://github.com/tompave/fun_with_flags/pull/151))\n* Add support for text (binary) primary key columns with the Ecto adapter. (Thanks [whatyouhide](https://github.com/whatyouhide) and [vinniefranco](https://github.com/vinniefranco), [pull/156](https://github.com/tompave/fun_with_flags/pull/156) and [pull/129](https://github.com/tompave/fun_with_flags/pull/129))","ref":"changelog.html#v1-11-0"},{"type":"extras","title":"v1.10.1 - Changelog","doc":"* Redis notifications adapter: minor internal changes to what data is passed to the supervisor in the child spec. ([pull/148](https://github.com/tompave/fun_with_flags/pull/148))\n* Redis notifications adapter: fix an issue that would arise when configuring both a Redis URL string and key-word options (this config API was introduced in v1.10.0): the config would work for the Redis persistence adapter, but not for the Redis notifications adapter. (Thanks [iamvery](https://github.com/iamvery), [pull/149](https://github.com/tompave/fun_with_flags/pull/149))","ref":"changelog.html#v1-10-1"},{"type":"extras","title":"v1.10.0 - Changelog","doc":"* Add support for Elixir 1.14. Drop support for Elixir 1.11. Elixir >= 1.12 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* Drop support for Erlang/OTP 22, and Erlang/OTP >= 23 is now required. Dropping support for older versions of Erlang/OTP simply means that this package is not tested with them in CI, and that no compatibility issues are considered bugs.\n* Ecto persistence adapter: FunWithFlags will now pass a custom option when using the [Ecto Repo Query API](https://hexdocs.pm/ecto/3.8.4/Ecto.Repo.html#query-api): `[fun_with_flags: true]`. This is done to make it easier to identify FunWithFlags queries when working with Ecto customization hooks, e.g. the [`Ecto.Repo.prepare_query/3` callback](https://hexdocs.pm/ecto/3.8.4/Ecto.Repo.html#c:prepare_query/3). (Thanks [SteffenDE](https://github.com/SteffenDE), [pull/143](https://github.com/tompave/fun_with_flags/pull/143))\n* Redis persistence adapter: added support to configure Redis with a `{\"redis URL\", [...kw opts]}` tuple, [as is supported in Redix itself](https://hexdocs.pm/redix/1.2.0/Redix.html#start_link/2). (Thanks [iamvery](https://github.com/iamvery), [pull/145](https://github.com/tompave/fun_with_flags/pull/145))","ref":"changelog.html#v1-10-0"},{"type":"extras","title":"v1.9.0 - Changelog","doc":"* Drop support for Elixir 1.10. Elixir >= 1.11 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* Relax supported versions of `postgrex` to allow `~> 0.16`.\n* Use [`Application.compile_env/3`](https://hexdocs.pm/elixir/1.13.3/Application.html#compile_env/3) to read the persistence config at compile time, which is used to configure the DB table name when using the Ecto persistence adapter (among other things). This fixes another instance of the issue where users of the package would change the config after compilation and observe unexpected inconsistencies and errors. ([pull/130](https://github.com/tompave/fun_with_flags/pull/130))\n* Redis adapters: add support to configure [Redis Sentinel](https://redis.io/docs/manual/sentinel/). Please see the [Redix docs](https://github.com/whatyouhide/redix/tree/v1.1.5#redis-sentinel) for more details. (Thanks [parkdoyeon](https://github.com/parkdoyeon), [pull/107](https://github.com/tompave/fun_with_flags/pull/107).)\n* More precise conditional checks when deciding whether Ecto files should be compiled. ([pull/140](https://github.com/tompave/fun_with_flags/pull/140))\n* Improved documentation for running the package in a custom supervision tree, when using releases. (Thanks [zaid](https://github.com/zaid), [pull/139](https://github.com/tompave/fun_with_flags/pull/139))","ref":"changelog.html#v1-9-0"},{"type":"extras","title":"v1.8.1 - Changelog","doc":"* Lock `postgrex` dependency to `< 0.16`. Version `0.16` requires Elixir 1.11 ([changelog](https://github.com/elixir-ecto/postgrex/blob/master/CHANGELOG.md#v0160-2022-01-23)) and it doesn't compile with Elixit 1.10, which FunWithFlags still supports.","ref":"changelog.html#v1-8-1"},{"type":"extras","title":"v1.8.0 - Changelog","doc":"* Add support for Elixir 1.13. Drop support for Elixir 1.9. Elixir >= 1.10 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* Removed all uses of [`defdelegate/2`](https://hexdocs.pm/elixir/1.13.0/Kernel.html#defdelegate/2). They caused some references to configured modules (that can change according to the config) to be reified at compile time, which lead to unexpected behaviour. They've been replaced with plain old function definitions that do the same job. (Thanks [connorlay](https://github.com/connorlay), [pull/111](https://github.com/tompave/fun_with_flags/pull/111).)\n* Local dev: Update the config for the library to use [`Config`](https://hexdocs.pm/elixir/1.13.0/Config.html) instead of the deprecated [`Mix.Config`](https://hexdocs.pm/mix/1.13.0/Mix.Config.html). For the avoidance of doubt: this has no effect when using the package in your projects, because the `config/*.exs` files are not present in the bundles downloaded from Hex.pm.\n* Use [`Application.compile_env/3`](https://hexdocs.pm/elixir/1.13.0/Application.html#compile_env/3) to read the cache configuration at compile-time, which is used to define a module attribute (therefore, set at compile-time). That part of the config is compiled into a module attribute for performance reasons, and it has been a long standing issue because users of the package would get confused by their config changes not being reflected in an already compiled application ([link to relevant section in previous version of the readme](https://github.com/tompave/fun_with_flags/tree/v1.7.0#configuration-changes-have-no-effect-in-mix_envdev)). Now, if the relevant configuration changes, users will get a [clear error](https://github.com/elixir-lang/elixir/blob/v1.10/CHANGELOG.md#tracking-of-compile-time-configuration).\n* Improve error handling in different layers of the package. From the persistence adapters all the way to the public functions of the top-level module. In practice, this means that some situations that would have caused a `MatchError` now instead will bubble up an error tuple. Most importantly, this does **not** affect the signature or behaviour of the `FunWithFlags.enabled?/2` function, which continues to return a simple boolean. ([pull/120](https://github.com/tompave/fun_with_flags/pull/120))\n* Typespec improvements. These include new typespecs for previously unspecced functions, amended typespecs for the new error tuples that are now bubbled up (see previous point) and fixed typespec that incorrectly ignored a returned error tuple. ([pull/120](https://github.com/tompave/fun_with_flags/pull/120))\n* The typespecs for the `FunWithFlags.Store.Persistence` Elixir behaviour have been updated (see previous point). Users of the package who implemented their own custom persistence adapters are encouraged to double-check that these respect the typespecs. ([pull/120](https://github.com/tompave/fun_with_flags/pull/120))","ref":"changelog.html#v1-8-0"},{"type":"extras","title":"v1.7.0 - Changelog","doc":"* Add support for Elixir 1.12. Drop support for Elixir 1.8. Elixir >= 1.9 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* Drop support for Erlang/OTP 21, and Erlang/OTP >= 22 is now required. Dropping support for older versions of Erlang/OTP simply means that this package is not tested with them in CI, and that no compatibility issues are considered bugs.\n* Added support for the Erlang [dialyzer](https://erlang.org/doc/man/dialyzer.html) (via the [dialyxir](https://hex.pm/packages/dialyxir) package).\n* Addressed all dialyzer warnings. Fixed some incorrect typespecs and simplified the implementation of some functions.\n* Miscellaneous documentation fixes and improvements. (Thanks [kianmeng](https://github.com/kianmeng), [pull/89](https://github.com/tompave/fun_with_flags/pull/89), [pull/90](https://github.com/tompave/fun_with_flags/pull/90) and [pull/112](https://github.com/tompave/fun_with_flags/pull/112).)\n* Documented the `FunWithFlags.Store.Cache` module, and its `Cache.flush/0` and `Cache.dump/0` functions. They're now part of the public API of the package.\n* Introduced a new `FunWithFlags.Supervisor` module to manage the supervision tree for the package. The supervision strategy and configuration are unchanged, and host applications don't need to do anything to upgrade. However, this module is part of the public API of the package and can be used to better control the start behaviour of FunWithFlags. This has also been documented [in a new section of the readme](https://github.com/tompave/fun_with_flags#application-start-behaviour).\n* Internal changes to stop using an undocumented feature of Elixir that will go away in future versions. This affects how the function to calculate Actor scores for the %-of-actors gate is invoked, but that's an internal change, so it won't affect users of the package unless they're using undocumented features. (Thanks [kelvinst](https://github.com/kelvinst), [pull/105](https://github.com/tompave/fun_with_flags/pull/105).)","ref":"changelog.html#v1-7-0"},{"type":"extras","title":"v1.6.0 - Changelog","doc":"* Add support for Elixir 1.11. Drop support for Elixir 1.7. Elixir >= 1.8 is now required. Dropping support for older versions of Elixir simply means that this package is no longer tested with them in CI, and that compatibility issues are not considered bugs.\n* More internal changes to not compile in the package configuration. Removed compile-time references to the Ecto repo and the Ecto table name. See the release notes for v1.5.1 (below) for more details on this type of changes.\n* Ecto and Postgres persistence: when updating percentage gates, use a flag-scoped advisory lock rather than locking the entire table. With the old system the entire table was locked when setting or changing any percentage gate, across all flags. With this change, the lock is scoped to one flag and the table is never fully locked.\n* Dev and test fixes to support Phoenix.PubSub on OTP 23 and Elixir >= 1.10.3. This was only an issue when working locally, and there should be no problems when using the previous version of the package in a host application.\n* Update Redix to 1.0. As [its changelog](https://github.com/whatyouhide/redix/blob/main/CHANGELOG.md#v100) says this doesn't introduce breaking changes, but it's a major version bump that should be documented here, as it will require changes in the host applications mix files.","ref":"changelog.html#v1-6-0"},{"type":"extras","title":"v1.5.1 - Changelog","doc":"* Internal changes to not compile in the persistence adapter config. This has no effect on the functionality of the package, but now the Ecto or Redis adapter configuration is not memoized anymore, and it can be changed with no need to recompile the package.","ref":"changelog.html#v1-5-1"},{"type":"extras","title":"v1.5.0 - Changelog","doc":"* Drop support for Elixir 1.6. Elixir >= 1.7 is now required.\n* Drop support for Erlang/OTP 20, and Erlang/OTP >= 21 is now required. An older Erlang/OTP might still work with older versions of Elixir, but Elixir 1.10 requires Erlang/OTP >= 21. Dropping support for older versions of Erlang/OTP simply means that this package is not tested with them in CI, and that no compatibility issues are considered bugs.\n* Upgrade Phoenix.PubSub dependency to 2.0. This provides compatibility with Phoenix 1.5.\n* Typespec improvements. (Thanks [skylerparr](https://github.com/skylerparr), [pull/57](https://github.com/tompave/fun_with_flags/pull/57))\n* Internal changes to how flag data is cached in the ETS table. This has no effect on the functionality of the package, with two exceptions. First, the `cache: [ttl: seconds]` config value is not memoized anymore and it can be changed without recompiling. Second, since the TTL is now stored with the ETS entries, old and new ETS data is not compatible; this is not an issue if you restart/rotate your application nodes/instances when deploying, but it will be an issue if you perform [hot code upgrades](https://hexdocs.pm/mix/1.9.4/Mix.Tasks.Release.html#module-hot-code-upgrades). In that case, you have to first empty the ETS table, for example with `FunWithFlags.Store.Cache.flush/0`.\n* New config option to set a custom name for the DB table when using the Ecto persistence adapter. (Thanks [BobbyMcWho](https://github.com/BobbyMcWho), [pull/64](https://github.com/tompave/fun_with_flags/pull/64) and [pull/77](https://github.com/tompave/fun_with_flags/pull/77))","ref":"changelog.html#v1-5-0"},{"type":"extras","title":"v1.4.1 - Changelog","doc":"* Typespec improvements. (Thanks [LostKobrakai](https://github.com/LostKobrakai), [pull/49](https://github.com/tompave/fun_with_flags/pull/49))\n* Improve Redis error handling for connection and command errors. (Thanks [chubarovNick](https://github.com/chubarovNick), [pull/54](https://github.com/tompave/fun_with_flags/pull/54))","ref":"changelog.html#v1-4-1"},{"type":"extras","title":"v1.4.0 - Changelog","doc":"This release focuses on making it easier to extend the package, for example with custom persistence adapters.\n\n* Define a [behaviour](https://hexdocs.pm/elixir/typespecs.html#behaviours) in the `FunWithFlags.Store.Persistence` module that can be implemented by custom persistence adapters. The builtin Redis and Ecto adapters now formally implement this new behaviour.\n* Refactor how cache-busting change notifications are published: move the logic out of the two builtin persistence adapters and into the level above them. While this is just an internal change, it narrows the responsibilities of the persistence adapters and simplifies implementing custom ones.\n* Update the supervision tree to use [Elixir v1.5 style child specs](https://github.com/elixir-lang/elixir/blob/v1.5/CHANGELOG.md#streamlined-child-specs).\n* Print a helpful error if a project is configured to use a persistence adapter without including its dependency packages. This mirrors what happens when the dependencies for a notifications adapter are missing.\n* Document the `Flag` and `Gate` types, previously private.\n* Redis persistence: relax Redix version lock to `~> 0.9`, which allows to use Redix `0.10`. It was previously locked to `~> 0.9.1` because of breaking changes in the last few Redix minor version releases, but going forward if it happens again it can be handled with a patch level release on FunWithFlags.","ref":"changelog.html#v1-4-0"},{"type":"extras","title":"v1.3.0 - Changelog","doc":"* Ecto persistence: added support for MySQL, via either `mariaex` or `myxql`. While both are working today, the test suite uses `myxql` because `ecto_sql` is going to deprecate the `mariaex` adapter in the future. (Thanks [stewart](https://github.com/stewart) for starting this work with [pull/41](https://github.com/tompave/fun_with_flags/pull/41) and for helping out on [pull/42](https://github.com/tompave/fun_with_flags/pull/42)).","ref":"changelog.html#v1-3-0"},{"type":"extras","title":"v1.2.1 - Changelog","doc":"* Fix invalid typespec that was causing compiler warnings on Elixir 1.8. (Thanks [asummers](https://github.com/asummers), [pull/34](https://github.com/tompave/fun_with_flags/pull/34))","ref":"changelog.html#v1-2-1"},{"type":"extras","title":"v1.2.0 - Changelog","doc":"* Redis persistence: upgrade to Redix 0.9, which deprecates Redix.PubSub. The pubsub capabilities are now part of the base Redix package. This means that FunWithFlags also needs to drop the dependency on Redix.PubSub.\n* Compatibility updates in the tests for Elixir 1.8.\n\nThere is no other change in this release, but this is a minor version bump because upgrading Redix and dropping Redix.PubSub will require applications to also update their dependencies.","ref":"changelog.html#v1-2-0"},{"type":"extras","title":"v1.1.0 - Changelog","doc":"* Drop support for Elixir 1.5. Elixir >= 1.6 is now required.\n* Drop support for OTP 19. OTP >= 20 is now required.\n* Update to Ecto 3 with the `ecto_sql` package.\n* Update to Redix 0.8 and Redix.PubSub 0.5.\n* Ecto persistence: add `NOT NULL` constraints to the table definition in the Ecto migration. This is not a breaking change: the constraints have been added because those values are never null anyway. If users of the library want to add them, they can do so by adding [this migration](https://github.com/tompave/fun_with_flags/blob/v1.1.0/priv/ecto_repo/migrations/00000000000001_ensure_columns_are_not_null.exs) to their projects.\n* Redis persistence: allow to configure the Redis URL with a system tuple to read it from an environment variable. (Thanks [seangeo](https://github.com/seangeo), [pull/29](https://github.com/tompave/fun_with_flags/pull/29))","ref":"changelog.html#v1-1-0"},{"type":"extras","title":"v1.0.0 - Changelog","doc":"This release introduces the last two gates that were initially planned and marks a milestone for the project. The API is now stable, and the project can graduate to `1.0.0`.\n\nThis release doesn't introduce any breaking change, however, and users of the library should be able to upgrade without problems. If you're also using [`FunWithFlags.UI`](https://github.com/tompave/fun_with_flags_ui) then make sure to also upgrade that to version `0.4.0`, which adds GUI support for the new features.\n\nNew Gates:\n\n* Percentage of time gate\n* Percentage of actors gate","ref":"changelog.html#v1-0-0"},{"type":"extras","title":"v0.11.0 - Changelog","doc":"* Add ability to clear the boolean gate only (useful for debugging).\n* Added `FunWithFlags.get_flag/1`, to retrieve a flag struct. Useful for debugging.\n* Internal improvements.","ref":"changelog.html#v0-11-0"},{"type":"extras","title":"v0.10.1 - Changelog","doc":"Improvements:\n\n* Ecto persistence: explicitly set the table primary key as an integer type. This improves the compatibility with Ecto repos where primary keys default to a binary type, e.g. UUID. (Thanks [coryodaniel](https://github.com/coryodaniel), [pull/23](https://github.com/tompave/fun_with_flags/pull/23))","ref":"changelog.html#v0-10-1"},{"type":"extras","title":"v0.10.0 - Changelog","doc":"Possibly Breaking Changes:\n\n* Allow binaries _and_ atoms as group gate names. Binaries are now preferred (atom group names are internally converted, stored and retrieved as binaries) and atoms are still allowed for retro-compatibility.  \nWhile calling `FunWithFlags.enable(:foo, for_group: :bar)` is still allowed and continues to work as before, this change will impact implementations of the `FunWithFlags.Group` protocol that assume\nthat the group name is passed as an atom.  \nTo safely upgrade, these implementations should be changed to work with the group names passed as a binary instead. See the [update to the protocol implementation used in the tests](https://github.com/tompave/fun_with_flags/pull/13/files#diff-8c1bcfc3d51e8d863953ac5b57f0da2b) for an example.\n\nOther changes:\n\n* Compatibility updates for Ecto 2.2 (dev env, was fine in prod)","ref":"changelog.html#v0-10-0"},{"type":"extras","title":"v0.9.2 - Changelog","doc":"Bug Fixes:\n\n* Fixed another issue with modules referencing Ecto.","ref":"changelog.html#v0-9-2"},{"type":"extras","title":"v0.9.1 - Changelog","doc":"Bug Fixes:\n\n* Fixed an issue with module referencing Ecto that was not wrapped in a `Code.ensure_loaded?` block, which prevented the library from being used in projects that did not include Ecto.","ref":"changelog.html#v0-9-1"},{"type":"extras","title":"v0.9.0 - Changelog","doc":"* Ecto persistence adapter. It's now possible to store flag data with Ecto instead of Redis; if used in conjunction with the Phoenix.PubSub adapter, it's possible to use this library in Phoenix without Redis.\n* The `redix` dependency is now optional.\n* Added optional `ecto` dependency.","ref":"changelog.html#v0-9-0"},{"type":"extras","title":"v0.8.1 - Changelog","doc":"* Mark the `redix_pubsub` dependency as optional.\n* Clearer error reporting for missing adapter dependencies.","ref":"changelog.html#v0-8-1"},{"type":"extras","title":"v0.8.0 - Changelog","doc":"New Features:\n\n* Added support for Phoenix.PubSub as an alternative transport for the cache busting notifications.\n* Added ability to enable the ETS cache but disable the cache-busting notifications, as it can be useful when running on a single node.\n\nOther changes:\n\n* Upgraded `redix` and `redix_pubsub` dependencies.\n* Internal project and supervision changes to better support different adapters.\n* Updated the Mix configuration options.\n* More rational test setup.","ref":"changelog.html#v0-8-0"},{"type":"extras","title":"v0.7.1 - Changelog","doc":"Bug fixes:\n\n* Resolved an issue with the PubSub connection process that would crash the entire supervision tree in case of abrupt disconnection from Redis. Ops!","ref":"changelog.html#v0-7-1"},{"type":"extras","title":"v0.7.0 - Changelog","doc":"New Features:\n\n* `FunWithFlags.all_flags/0` is now public and documented.\n* Added `FunWithFlags.all_flag_names/0`, public and documented.\n* Added proper log statements via the Elixir `Logger`. Setting the log level to `debug` will print cache busting info, for example.\n\nInternal changes:\n\n* Updated the `redix` and `redix_pubsub` dependencies.\n* Extracted the private persistence and notifications modules into a redis-specific namespace. Added a config option to customize the adapters on startup, and an internal API that allows to develop other adapters to use alternative persistence and notification layers. The persistence adapters are responsible for declaring their own notifications module (if any). The provided default ones keep using Redis, and they work in tandem. At the moment, no official support for other adapters is planned, `redix` stays as a dependency.","ref":"changelog.html#v0-7-0"},{"type":"extras","title":"v0.6.0 - Changelog","doc":"New features:\n\n* Added `FunWithFlags.clear/2` to delete a specific gate or an entire flag. This is useful if you don't need an actor or group override and want to use the default boolean rule instead. Clearing a flag or a gate uses the same PubSub cache busting functionality used when updating a flag.\n* Added `FunWithFlags.all_flags/0`, to return a list of all the flags stored in Redis. Undocumented because it's meant to build a GUI.","ref":"changelog.html#v0-6-0"},{"type":"extras","title":"v0.5.0 - Changelog","doc":"Added the `Group` protocol and group gates. It's now possible to enable or disable a flag for a group name and implement `Group` for types and structs that should belong to groups.","ref":"changelog.html#v0-5-0"},{"type":"extras","title":"v0.4.0 - Changelog","doc":"Added the `Actor` protocol and actor gates. It's now possible to implement `Actor` for some type or struct and then enable or disable a flag for some specific values.","ref":"changelog.html#v0-4-0"},{"type":"extras","title":"v0.3.0 - Changelog","doc":"Always raise exceptions if Redis becomes unavailable _and_ there is no cached value (expired or not). This means that both with or without cache, failures to load a flag's data will never be silently converted to the flag being disabled.","ref":"changelog.html#v0-3-0"},{"type":"extras","title":"v0.2.0 - Changelog","doc":"Significant internal rewrite: use structures instead of raw booleans. Given the scope of the changes and the fact that this is still a `0.x` release, bump the version number.","ref":"changelog.html#v0-2-0"},{"type":"extras","title":"v0.1.1 - Changelog","doc":"* Enhancements\n    * Treat cache misses and expired cached values differently. If Redis becomes unavailable, and an expired value is available in the cache, use the expired value even though normally it would be discarded.","ref":"changelog.html#v0-1-1"},{"type":"extras","title":"v0.1.0 - Changelog","doc":"First usable release with the a stable initial feature set.\n\n* Simple boolean flags\n* Elixir API to enable, disable and query the flags\n* Supervision tree, embeddable in host applications\n* Persistence in Redis\n* ETS cache\n    * Cache busting with TTLs\n    * Cache busting with Redis PubSub\n* Resistant to Redis connection issues if the values are cached\n* Option to disable the ETS cache (Redis-only mode)","ref":"changelog.html#v0-1-0"},{"type":"extras","title":"v0.0.x - Changelog","doc":"Unstable releases.","ref":"changelog.html#v0-0-x"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}